@@ -1,6 +1,7 @@
 package main
 
 import (
+	"crypto/x509"
 	stderrors "errors"
 	"testing"
 	"time"
@@ -11,7 +12,6 @@ import (
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime/schema"
 	corev1client "k8s.io/client-go/kubernetes/typed/core/v1"
-	"k8s.io/client-go/util/certificate/csr"
 
 	machinev1beta1 "github.com/openshift/cluster-api/pkg/apis/machine/v1beta1"
 )
@@ -231,7 +231,7 @@ Certificate Request:
         Subject: O = system:nodes, CN = system:node:monkey
 ...
         Requested Extensions:
-            X509v3 Subject Alternative Name: 
+            X509v3 Subject Alternative Name:
                 DNS:banana
 ...
 -----BEGIN CERTIFICATE REQUEST-----
@@ -308,6 +308,7 @@ func Test_authorizeCSR(t *testing.T) {
 		nodeErr  error
 		req      *certificatesv1beta1.CertificateSigningRequest
 		csr      string
+		ca       *x509.CertPool
 	}
 	tests := []struct {
 		name    string
@@ -408,7 +409,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: goodCSR,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"test\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "only-node-prefix",
@@ -482,29 +483,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: goodCSR,
 			},
-			wantErr: "No target machine",
-		},
-		{
-			name: "no-machine-status",
-			args: args{
-				machines: nil,
-				req: &certificatesv1beta1.CertificateSigningRequest{
-					Spec: certificatesv1beta1.CertificateSigningRequestSpec{
-						Usages: []certificatesv1beta1.KeyUsage{
-							certificatesv1beta1.UsageDigitalSignature,
-							certificatesv1beta1.UsageKeyEncipherment,
-							certificatesv1beta1.UsageServerAuth,
-						},
-						Username: "system:node:test",
-						Groups: []string{
-							"system:authenticated",
-							"system:nodes",
-						},
-					},
-				},
-				csr: goodCSR,
-			},
-			wantErr: "Invalid request",
+			wantErr: "No target machine for node \"test\"",
 		},
 		{
 			name: "missing-groups-1",
@@ -695,7 +674,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: goodCSR,
 			},
-			wantErr: "Not in system:authenticated",
+			wantErr: "map[\"system:authenticated\":{} \"system:foo-bar\":{}] not in \"system:authenticated\" and \"system:nodes\"",
 		},
 		{
 			name: "usages-missing",
@@ -1029,7 +1008,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: noGroup,
 			},
-			wantErr: "Organization doesn't include system:nodes",
+			wantErr: "Organization [] doesn't include system:nodes",
 		},
 		{
 			name: "csr-extra-addr",
@@ -1256,7 +1235,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientExtraO,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client with DNS",
@@ -1292,7 +1271,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientWithDNS,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client good but extra usage",
@@ -1329,7 +1308,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client good but wrong usage",
@@ -1365,7 +1344,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client good but missing usage",
@@ -1400,7 +1379,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client good but wrong CN",
@@ -1436,7 +1415,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientWrongCN,
 			},
-			wantErr: "Doesn't match expected prefix",
+			wantErr: "\"system:serviceaccount:openshift-machine-config-operator:node-bootstrapper\" doesn't match expected prefix: \"system:node:\"",
 		},
 		{
 			name: "client good but wrong user",
@@ -1473,7 +1452,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "CSR green for node client cert has wrong user",
+			wantErr: "CSR green for node client cert has wrong user system:serviceaccount:openshift-machine-config-operator:node-bootstrapper-not or groups map[system:authenticated:{} system:serviceaccounts:{} system:serviceaccounts:openshift-machine-config-operator:{}]",
 		},
 		{
 			name: "client good but wrong user group",
@@ -1511,7 +1490,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "CSR blue for node client cert has wrong user",
+			wantErr: "CSR blue for node client cert has wrong user system:serviceaccount:openshift-machine-config-operator:node-bootstrapper or groups map[extra-group:{} system:authenticated:{} system:serviceaccounts:{} system:serviceaccounts:openshift-machine-config-operator:{}]",
 		},
 		{
 			name: "client good but empty name",
@@ -1904,7 +1883,7 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "CSR purple creation time 2000-01-01 02:32:00 +0000 UTC not in range (2000-01-01 02:32:50 +0000 UTC, 2000-01-01 02:43:00 +0000 UTC)",
+			wantErr: "CSR purple creation time 2000-01-01 02:32:00 +0000 UTC not in range (2000-01-01 02:32:50 +0000 UTC, 2000-01-01 04:33:00 +0000 UTC)",
 		},
 		{
 			name: "client good but CSR too late",
@@ -1939,7 +1918,7 @@ func Test_authorizeCSR(t *testing.T) {
 				req: &certificatesv1beta1.CertificateSigningRequest{
 					ObjectMeta: metav1.ObjectMeta{
 						Name:              "red",
-						CreationTimestamp: creationTimestamp(15 * time.Minute),
+						CreationTimestamp: creationTimestamp(25 * time.Hour),
 					},
 					Spec: certificatesv1beta1.CertificateSigningRequestSpec{
 						Usages: []certificatesv1beta1.KeyUsage{
@@ -1957,25 +1936,122 @@ func Test_authorizeCSR(t *testing.T) {
 				},
 				csr: clientGood,
 			},
-			wantErr: "CSR red creation time 2000-01-01 02:45:00 +0000 UTC not in range (2000-01-01 02:32:50 +0000 UTC, 2000-01-01 02:43:00 +0000 UTC)",
+			wantErr: "CSR red creation time 2000-01-02 03:30:00 +0000 UTC not in range (2000-01-01 02:32:50 +0000 UTC, 2000-01-01 04:33:00 +0000 UTC)",
 		},
 	}
 	for _, tt := range tests {
 		t.Run(tt.name, func(t *testing.T) {
 			tt.args.req.Spec.Request = []byte(tt.args.csr)
-			parsedCSR, err := csr.ParseCSR(tt.args.req)
+			parsedCSR, err := parseCSR(tt.args.req)
 			if err != nil {
 				t.Fatal(err)
 			}
 			nodes := &testNode{t: t, name: tt.args.nodeName, err: tt.args.nodeErr}
 
-			if err := authorizeCSR(tt.args.config, tt.args.machines, nodes, tt.args.req, parsedCSR); errString(err) != tt.wantErr {
+			if err := authorizeCSR(tt.args.config, tt.args.machines, nodes, tt.args.req, parsedCSR, tt.args.ca); errString(err) != tt.wantErr {
 				t.Errorf("authorizeCSR() error = %v, wantErr %s", err, tt.wantErr)
 			}
 		})
 	}
 }
 
+func TestAuthorizeServingRenewal(t *testing.T) {
+	// TODO(bison): Need actual coverage here.
+	tests := []struct {
+		name        string
+		nodeName    string
+		csr         *x509.CertificateRequest
+		currentCert *x509.Certificate
+		ca          *x509.CertPool
+		wantErr     string
+	}{
+		{
+			name:     "missing args",
+			nodeName: "panda",
+			wantErr:  "CSR, serving cert, or CA not provided",
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			err := authorizeServingRenewal(
+				tt.nodeName,
+				tt.csr,
+				tt.currentCert,
+				tt.ca,
+			)
+
+			if errString(err) != tt.wantErr {
+				t.Errorf("got: %v, want: %s", err, tt.wantErr)
+			}
+		})
+	}
+}
+
+func TestNodeInternalIP(t *testing.T) {
+	tests := []struct {
+		name    string
+		node    *corev1.Node
+		wantIP  string
+		wantErr string
+	}{
+		{
+			name: "no addresses",
+			node: &corev1.Node{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "no-addresses",
+				},
+				Status: corev1.NodeStatus{
+					Addresses: []corev1.NodeAddress{},
+				},
+			},
+			wantErr: "node no-addresses has no internal addresses",
+		},
+		{
+			name: "no internal ip",
+			node: &corev1.Node{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "no-internal-ip",
+				},
+				Status: corev1.NodeStatus{
+					Addresses: []corev1.NodeAddress{
+						{Type: corev1.NodeHostName, Address: "host.example.com"},
+					},
+				},
+			},
+			wantErr: "node no-internal-ip has no internal addresses",
+		},
+		{
+			name: "has internal ip",
+			node: &corev1.Node{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "has-internal-ip",
+				},
+				Status: corev1.NodeStatus{
+					Addresses: []corev1.NodeAddress{
+						{Type: corev1.NodeInternalIP, Address: "10.0.0.1"},
+					},
+				},
+			},
+			wantIP: "10.0.0.1",
+		},
+	}
+
+	for _, tt := range tests {
+		t.Run(tt.name, func(t *testing.T) {
+			ip, err := nodeInternalIP(tt.node)
+
+			if errString(err) != tt.wantErr {
+				t.Errorf("got: %v, want: %s", err, tt.wantErr)
+			}
+
+			if ip != tt.wantIP {
+				t.Errorf("got: %v, want: %s", err, tt.wantIP)
+			}
+		})
+	}
+}
+
 func errString(err error) string {
 	if err == nil {
 		return ""
@@ -2007,3 +2083,20 @@ func (n *testNode) Get(name string, _ metav1.GetOptions) (*corev1.Node, error) {
 
 	return nil, n.err
 }
+
+func TestGetMaxPending(t *testing.T) {
+	ml := []machinev1beta1.Machine{
+		{
+			Status: machinev1beta1.MachineStatus{},
+		},
+		{
+			Status: machinev1beta1.MachineStatus{},
+		},
+	}
+
+	res := getMaxPending(ml)
+	expected := len(ml) + maxPendingCSRs
+	if res != expected {
+		t.Errorf("getMaxPending returned incorrect value: %v, expect: %v", res, expected)
+	}
+}
